# The GPLv3 License (GPLv3)

# Copyright (c) 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from importlib.metadata import distribution, version

import pygame as pg

import darwinio.constants as constants
import darwinio.graphical_sim as gsim
import darwinio.stats as statistics


def main(resolution: tuple[int, int], fps: int, world_size: tuple[int, int]):
    """
    The main function that runs the game.

    Args:
    -----
    resolution: The resolution of the game screen.

    fps: The desired frame rate of the game.

    world_size: The size of the world for simulation.
    """

    # Initialize Pygame
    pg.init()
    screen = pg.display.set_mode(resolution, pg.SCALED | pg.RESIZABLE)
    pg.display.set_caption(f'darwinio v{version("darwinio")}')
    with gsim.get_asset_path("art", "eubacteria_BGA.png") as path:
        pg.display.set_icon(pg.image.load(path))
    clock = pg.time.Clock()

    # music
    music_playing = True
    with gsim.get_asset_path("audio", "Darwinio.ogg") as path:
        pg.mixer.music.load(path)
    pg.mixer.music.set_volume(1)
    pg.mixer.music.play()

    world = gsim.World(world_size, initial_temp_avg=45)
    stats = statistics.StatisticsCollector(
        ["Population", "Food", "Temperature", "Reproduction Ratio"]
    )

    # Create the states
    title = gsim.TitleScreen(
        screen, constants.TITLE_ASCII_ART, f'v{version("darwinio")}', 1
    )
    disclaimer = gsim.HeadingTextScreen(screen, "DISCLAIMER", constants.DISCLAIMER, 2)
    license_notice = gsim.TextScreen(screen, constants.LICENSE_NOTICE, 3)
    organism_build = gsim.OrganismSelection(screen, world, 5)
    init_help_screen = gsim.TextScreen(screen, constants.HELP, 4)
    help_screen = gsim.TextScreen(screen, constants.HELP, 6)
    distribution_painting = gsim.DistributionPainting(screen, world, 4)

    # get images for the organisms.
    image_names: list[str] = [
        "archaebacteria_halophile.png",
        "archaebacteria_methanogen.png",
        "archaebacteria_thermoacidophile.png",
        "eubacteria_BGA.png",
        "eubacteria_mycoplasma.png",
        "eubacteroa_chemosynthetic.png",
        "fungi-ascomycetes.png",
        "fungi-basidiomycetes.png",
        "fungi-deuteromycetes.png",
        "fungi-phycomycetes.png",
        "metaphyta-algae.png",
        "metaphyta-angiospermae.png",
        "metaphyta-bryophyta.png",
        "metaphyta-gymnospermae.png",
        "metaphyta-pterdiophyta.png",
        "protista_dinoflagellate.png",
        "protista_euglena.png",
        "protista_protozoan.png",
        "protista_slimemould.png",
    ]
    images: list[pg.Surface] = []
    for file_name in image_names:
        with gsim.get_asset_path("art", file_name) as path:
            images.append(
                pg.transform.scale(pg.image.load(path).convert_alpha(), (64, 64))
            )

    main_game = gsim.Simulation(screen, world, stats, images)

    # Create the state machine
    statemachine = gsim.StateMachine(
        [
            title,
            disclaimer,
            license_notice,
            init_help_screen,
            organism_build,
            main_game,
            help_screen,
            distribution_painting,
        ]
    )

    while True:
        time_delta = clock.tick(fps) / 1000.0

        events = pg.event.get()
        for event in events:
            if event.type == pg.QUIT:
                pg.quit()
                raise SystemExit
            if event.type == pg.KEYUP:
                if event.key == pg.K_F11:
                    pg.display.toggle_fullscreen()
                if event.key == pg.K_ESCAPE:
                    pg.quit()
                    raise SystemExit
                if event.key == pg.K_m:
                    if music_playing:
                        pg.mixer.music.pause()
                        music_playing = not music_playing
                    else:
                        pg.mixer.music.unpause()
                        music_playing = not music_playing
                if event.mod & pg.KMOD_LSHIFT and event.key == pg.K_h:
                    prev_index: int = statemachine.state_index
                    statemachine.state_index = 5
                    help_screen.next_state_index = prev_index

        screen.fill("#000000")
        statemachine.run_state(events, time_delta)
        pg.display.flip()


def cli():
    main((1000, 750), 60, (50, 50))


if __name__ == "__main__":
    cli()



# The GPLv3 License (GPLv3)

# Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A module for representing a World with an array representing each
layer.

Classes:
---------
World: Represents a world in which organisms and food are distributed
across a canvas.

Functions:
----------
get_neighbour_cells: the values of neighbouring cells around a given
coordinate in a distribution.

get_distribution_population: Get the number of "truthy" values in the
distribution.

get_feasible_position: Finds a feasible position given a current
position, preferred position, and a distribution.

get_points_between_2_points: Return an array of coordinates of points
that lie on the line between two given points.
"""

from __future__ import annotations

import random
from typing import Union, cast

import numpy as np

import darwinio.organism as org


class Distribution:
    def __init__(self, data: np.ndarray) -> None:
        self.data: np.ndarray = data

    def get_population(self) -> int:
        """Get the number of "truthy" values in the distribution.

        Args:
        -----
        distribuion: A 2D array of values representing a distribution of some kind.
        """
        return np.count_nonzero(self.data)

    def get_neighbour_cells(self, coordinates: tuple[int, int]) -> np.ndarray:
        """Return the values of neighbouring cells around a given coordinate in a
        distribution.

        Args:
        -----
        coordinates: The (x, y) coordinate for which to retrieve neighbouring
        cells.

        Returns:
        --------
        np.ndarray: A 2D array containing the values of neighbouring cells around
        the given coordinates. Specifically, this array contains a 3x3 subset (not
        always) of ` distribution` centered around the given coordinates.
        """
        x, y = coordinates
        rows, cols = np.shape(self.data)
        return self.data[
            np.clip(x - 1, 0, rows) : np.clip(x + 1, 0, rows) + 1,
            np.clip(y - 1, 0, cols) : np.clip(y + 1, 0, cols) + 1,
        ]

    def get_feasible_position(
        self,
        current_position: tuple[int, int],
        preferred_position: tuple[int, int],
    ) -> tuple[int, int]:
        """Finds a feasible position given a current position, preferred
        position, and a distribution.

        Args:
        ----
        current_position: A tuple containing the x and y coordinates of the current
        position.

        preferred_position: A tuple containing the x and y coordinates of the
        preferred position.

        distribution: A 2D numpy array representing the distribution of feasible
        positions.

        Returns:
        -------
        A tuple containing the x and y coordinates of a feasible position. If there
        are no feasible positions between the current and preferred positions,
        returns the preferred position if it is feasible, otherwise returns the
        current position.
        """
        x, y = self.data.shape
        possible_positions: np.ndarray = get_points_between_2_points(
            current_position, preferred_position
        )[:-1]

        for index, position in enumerate(possible_positions):
            row, column = tuple(position)
            if self.data[np.clip(column, 0, y - 1)][np.clip(row, 0, x - 1)]:
                return cast(
                    tuple[int, int],
                    tuple(
                        np.clip(
                            possible_positions[index - 1 if index != 0 else index][p],
                            0,
                            (y, x)[p] - 1,
                        )
                        for p in range(2)
                    ),
                )
        return cast(
            tuple[int, int],
            tuple(
                np.array(
                    [np.clip(preferred_position[p], 0, (x, y)[p] - 1) for p in range(2)]
                ).astype(int)
            ),
        )

    @classmethod
    def generate(
        cls,
        size: tuple[int, int],
        loc: int,
        scale: int,
    ) -> Distribution:
        """
        Generate a 2D numpy array of random values sampled from a
        normal distribution.

        Args:
        -----
        loc: An integer representing the mean of the normal
        distribution.

        scale: An integer representing the standard deviation of the
        normal distribution.

        size: A tuple of two integers representing the dimensions of
        the distribution.

        Returns:
        --------
        A 2D numpy array with the shape canvas_size where each element
        is a random value sampled from a normal distribution with the
        specified mean and standard deviation.
        """
        return cls(np.random.normal(loc=loc, scale=scale, size=size).astype(int))


class OrganismDistribution(Distribution):
    def get_reproduction_ratio(self) -> float:
        """
        Returns:
        --------
        The 'reproduction ratio', which is basically the
        number of asexuals per sexuals.
        """
        asexuals: int = 0
        sexuals: int = 0
        for row in self.data:
            for organism in row:
                if organism is not None:
                    if organism.genome_array[2] == 0:
                        asexuals += 1
                    else:
                        sexuals += 1
        return asexuals / sexuals if sexuals != 0 else np.NaN

    @classmethod
    def generate(
        cls,
        size: tuple[int, int],
        weights: tuple[float, float] = (0.1, 0.9),
    ) -> OrganismDistribution:
        """
        Generate a random organism distribution.
        """
        return cls(
            np.array(
                [
                    [
                        random.choices(
                            (
                                org.Organism.random(),
                                None,
                            ),
                            weights=weights,
                            k=1,
                        )[0]
                        for _ in range(size[1])
                    ]
                    for _ in range(size[0])
                ],
                dtype=object,
            )
        )


class World:
    """Represents a world in which organisms and food are distributed
    across a canvas.

    Attributes:
    -----------
    canvas_size: A tuple of two integers representing the dimensions
    of the canvas.

    mutation_factor: A value between 0 and 1 (inclusive) representing
    the probability of a mutation occurring in the offspring's genome.

    food_distribution: A numpy array of random integers between 0 and
    5000 of size `canvas_size`.

    organism_distribution: A 2D list of organisms and `None` values of
    size `canvas_size`.
    """

    def __init__(
        self,
        canvas_size: tuple[int, int],
        mutation_factor: float = 0.3,
        initial_food_avg: int = 500,
        initial_temp_avg: int = 315,
    ):
        """Initializes a new World instance.

        Args:
        ------
        canvas_size: A tuple of two integers representing the dimensions of the
        canvas.

        mutation_factor: A value between 0 and 1 (inclusive) representing the
        probability of a mutation occurring in the offspring's genome.

        initial_food_distribution: The range of values which will be found in
        the food variation.

        initial_temp_distribution: The range of values which will be found in
        the temp variation.
        """

        self.canvas_size: tuple[int, int] = canvas_size
        self.mutation_factor: float = mutation_factor

        self.food_distribution: Distribution = Distribution.generate(
            self.canvas_size,
            initial_food_avg,
            100,
        )

        self.temp_distribution: Distribution = Distribution.generate(
            self.canvas_size, initial_temp_avg, 50
        )
        self.organism_distribution: OrganismDistribution = (
            OrganismDistribution.generate(self.canvas_size)
        )

    def update_state(self) -> None:
        """Update the state of the canvas.

        Note:
        -----
        Updates the state of the world by iterating over each organism and
        updating its position based on its neural network's output. If another
        organism is not present at its current position after updating, it is
        removed from the current position and added to the new position. It
        also considers the direction of food around it. Then it allows the
        organism to reproduce if it has access to 2x the amount of food.
        """

        rows, cols = self.canvas_size

        for i in range(rows):
            for j in range(cols):
                organism: org.Organism = self.organism_distribution.data[i][j]

                # check if there is an organism at the current location
                if organism is not None:
                    temp_range = get_integer_neighbors(organism.genome_array[0], 150)
                    food_value: int = self.food_distribution.data[i][j]

                    # name the conditions
                    has_enough_food: bool = food_value >= organism.genome_array[1]
                    is_in_ideal_temp: bool = (
                        self.temp_distribution.data[i][j] in temp_range
                    )
                    has_enough_food_for_reprod: bool = (
                        food_value >= 2 * organism.genome_array[1]
                    )
                    is_in_ideal_temp_for_reprod: bool = (
                        self.temp_distribution.data[i][j] in temp_range
                    )

                    if has_enough_food and is_in_ideal_temp:
                        self.food_distribution.data[i][j] -= organism.genome_array[1]
                        self.move(organism, (i, j))

                    if has_enough_food_for_reprod and is_in_ideal_temp_for_reprod:
                        self.reproduce(organism, (i, j))

                    # if food is not available kill it and derive some food
                    # from its dead body.
                    else:
                        self.food_distribution.data[i][j] += (
                            organism.genome_array[1] // 10
                        )
                        self.organism_distribution.data[i][j] = None

    def move(self, organism: org.Organism, current_position: tuple[int, int]):
        """Move the organism to a new position based on the current position
        and environmental conditions.

        Args:
        ------
        organism: An instance of the Organism class representing the organism
        to be moved.

        current_position: A tuple of two integers representing the current
        position of the organism.
        """

        i, j = current_position
        neighbour_cells_food_dist: np.ndarray = (
            self.food_distribution.get_neighbour_cells((i, j))
        )
        neighbour_cells_temp_dist: np.ndarray = (
            self.temp_distribution.get_neighbour_cells((i, j))
        )

        food_direction = int(
            np.argmax(neighbour_cells_food_dist)
            if np.size(neighbour_cells_food_dist.flatten())
            else -1
        )
        temp_direction = int(
            np.argmax(neighbour_cells_temp_dist)
            if np.size(neighbour_cells_temp_dist.flatten())
            else -1
        )

        new_x, new_y = (
            organism.neural_network.run_neural_network(
                np.array((food_direction, temp_direction))
            )
        ).astype(int)

        new_coordinates: tuple = self.organism_distribution.get_feasible_position(
            (i, j),
            (i + new_x, j + new_y),
        )

        # move the organism
        self.organism_distribution.data[i][j] = None
        self.organism_distribution.data[new_coordinates[0]][
            new_coordinates[1]
        ] = organism

    def reproduce(self, organism: org.Organism, current_position: tuple[int, int]):
        """Reproduce the organism at the current position.

        Args:
        ------
        organism: An instance of the Organism class representing the organism
        to be reproduced.

        current_position: A tuple of two integers representing the current
        position of the organism.
        """
        i, j = current_position
        prefered_position: tuple[int, int] = cast(
            tuple[int, int],
            tuple((i, j)[p] + np.random.choice((-1, 1)) for p in range(2)),
        )
        x, y = self.organism_distribution.get_feasible_position(
            (i, j),
            prefered_position,
        )

        offspring: Union[org.Organism, None]

        # asexual
        if organism.genome_array[2] == 0:
            offspring = org.reproduce(organism, organism, 0.3)
        # sexual
        else:
            partner: Union[org.Organism, None] = None
            neighbour_cells: np.ndarray = (
                self.organism_distribution.get_neighbour_cells((i, j))
            )
            for row in neighbour_cells:
                for other_organism in row:
                    if other_organism:
                        partner = other_organism
                        break
            if partner is not None:
                offspring = org.reproduce(organism, partner, 0.3)
            else:
                offspring = None

            if offspring:
                self.organism_distribution.data[y][x] = offspring


def get_points_between_2_points(
    point_1: tuple[int, int], point_2: tuple[int, int]
) -> np.ndarray:
    """Return an array of coordinates of points that lie on the line between
    two given points.

    Args:
    -----
    point_1: A tuple of two integers that represent the (x, y) coordinates of
    the first point.

    point_2: A tuple of two integers that represent the (x, y) coordinates of
    the second point.

    Returns:
    -----
    np.ndarray: An array of coordinates of the points that lie on the line
    between the two given points.(inclusive of point1 and point2)

    Note:
    -----
    The function calculates the coordinates of the points that lie on the line
    between the two input points, and returns an array of these coordinates. We
    first get the number of possible points integer points. And then finds such
    points such that they are linearly placed. And then combines them.
    """

    x_1, y_1 = np.array(point_1).astype(int)
    x_2, y_2 = np.array(point_2).astype(int)

    delta_y: int = y_2 - y_1
    delta_x: int = x_2 - x_1
    no_of_points: int = np.gcd(delta_y, delta_x) + 1

    x_coords: np.ndarray = np.linspace(x_1, x_2, no_of_points, dtype=int)
    y_coords: np.ndarray = np.linspace(y_1, y_2, no_of_points, dtype=int)

    # Combine the x and y coordinates into a single array
    points: np.ndarray = np.column_stack((x_coords, y_coords))

    return points


def get_integer_neighbors(value: int, radius: int) -> np.ndarray:
    """Get integers around a particular integer."""
    return np.arange(value - radius, value + radius + 1)



# The GPLv3 License (GPLv3)

# Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Organism class and related stuff.

Classes:
--------
Organism: A class representing an organism.

Functions:
--------
reproduce: Generate offspring of the two Organisms.
"""

from __future__ import annotations

import random

import numpy as np

import darwinio.brain as brn
import darwinio.genome as gn


class Organism:
    """A class representing an organism.

    Class Attributes:
    ---------
    temp_range: Range of temperature values for the organism's adaptation.

    energy_range: Range of energy values for the organism's energy capacity.

    reproductive_types: Range of reproductive type values for the organism's
    reproductive strategy.

    Attributes:
    ---------
    genome_array: A Numpy array representing the organism's genome.

    neural_network: A neural network generated from the genome of the organism
    """

    temp_range: tuple[int, int] = (30, 150)
    energy_range: tuple[int, int] = (100, 1000)
    reproductive_types: tuple[int, int] = (0, 1 + 1)

    def __init__(
        self,
        genome_array: np.ndarray,
    ) -> None:
        """Initializes an instance of the Organism class.

        Args:
        -----
        genome_array: A Numpy array representing the organism's genome.
        """

        self.genome_array: np.ndarray = genome_array

        # assign a neural_network generated from the genome
        neural_structure = np.array([2, 2])
        weights: np.ndarray = brn.create_weights(self.genome_array, neural_structure)
        self.neural_network = brn.NeuralNetwork(weights, neural_structure)

        # range
        self.temp_range: tuple[int, int] = (30, 150)
        self.energy_range: tuple[int, int] = (100, 1000)
        self.reproductive_types: tuple[int, int] = (0, 1 + 1)

    def get_tier(self) -> float:
        """Return the classification tier."""
        return np.sum(self.genome_array[:3]) / (
            self.temp_range[1] + self.energy_range[1] + self.reproductive_types[1]
        )

    @classmethod
    def random(cls) -> Organism:
        """Generate a random organism.

        Returns:
        ---------
        Organism: A random instance of the Organism class.
        """
        characters: np.ndarray = np.array(
            (
                random.randint(*sorted(cls.temp_range)),
                random.randint(*sorted(cls.energy_range)),
                random.randint(*sorted(cls.reproductive_types)),
            )
        )
        return cls(characters)


def reproduce(
    parent_1: Organism, parent_2: Organism, mutation_factor: float
) -> Organism:
    """Generate offspring of the two Organisms.

    Args:
    -----
    parent_1(np.ndarray): One of the parent Organisms

    parent_2(np.ndarray): One of the parent Organisms

    mutation_factor(float): A value between 0 and 1 (inclusive) representing
    the probability of a mutation occurring in the offspring's genome.

    Returns:
    ---------
    offspring: Child of the parents.
    """
    offspring_genome: np.ndarray = gn.generate_offspring_genome(
        parent_1.genome_array, parent_2.genome_array, mutation_factor
    )
    return Organism(offspring_genome)



# The GPLv3 License (GPLv3)

# Copyright (c) 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import copy
from importlib.metadata import Distribution
import threading
from importlib.resources import as_file, files
from typing import Any, Union, Tuple
import math

import pygame as pg
import pygame_gui as pgui
import numpy as np

import darwinio.distribution as dist
import darwinio.organism as org
import darwinio.genome as gn
import darwinio.stats as statistics


class World(dist.World):
    """Represents a world in which organisms and food are distributed across a
    canvas.

    An additional method to render the world.
    """

    def render(self, surface: pg.Surface, images: list[pg.Surface]):
        """
        Renders the organisms on the given surface using the provided image.

        Args:
        -----
        surface: The surface on which the organisms will be rendered.

        images: The images representing organisms.
        """
        organisms = self.organism_distribution.data
        for y, row in enumerate(organisms):
            for x, organism in enumerate(row):
                if organism is not None:
                    image = images[
                        np.clip(
                            int(organism.get_tier() * len(images)), 0, len(images) - 1
                        )
                    ]
                    color = pg.Color(f"#{gn.array2hex(organism.genome_array)[-6:]}")
                    tinted_image: pg.Surface = tint(image, color)
                    surface.blit(tinted_image, (x * 64, y * 64))


def render_np_2d_array(array: np.ndarray, surface: pg.Surface):
    """
    Renders the NumPy array on the given surface.

    Args:
    -----
    array: The NumPy array to be rendered.

    surface: The surface on which the NumPy array will be rendered.
    """
    color_pixel_size = size_x, size_y = tuple(
        surface.get_size()[_] / array.shape[_] for _ in range(2)
    )
    color_pixel: pg.Surface = pg.Surface(color_pixel_size)
    max_value: int = array.max()
    min_value: int = array.min()
    for y, row in enumerate(array):
        for x, datapoint in enumerate(row):
            color_brightness: int = round(
                255 * (datapoint - min_value) / (max_value - min_value)
                if max_value - min_value != 0
                else 0
            )

            # https://krazydad.com/tutorials/makecolors.php

            r = round(math.sin(0.01 * color_brightness + 0 * math.pi / 3) * 127 + 128)
            g = round(math.sin(0.01 * color_brightness + 1 * math.pi / 3) * 127 + 128)
            b = round(math.sin(0.01 * color_brightness + 2 * math.pi / 3) * 127 + 128)

            color: Tuple[int, int, int] = (r, g, b)

            color_pixel.fill(color)
            surface.blit(color_pixel, (size_x * x, size_y * y))


class State:
    """
    Represents a game
    state.

    Attributes:
    -----------
    manager: The UI manager for the state.

    surface: The surface on which the state is rendered.

    next_state_index: The index of the next state
    """

    def __init__(
        self,
        surface: pg.Surface,
        manager_size: tuple[int, int],
        next_state_index: Union[int, None],
    ):
        """
        Args:
        ------
        surface: The surface on which the state will be rendered.

        initial_manager_size: The initial size of the UI manager.

        next_state_index: The index of the next state
        """
        self.next_state_index: Union[int, None] = next_state_index
        self.manager = pgui.UIManager(manager_size)
        self.surface: pg.Surface = surface

    def render(self):
        """Render the state."""
        self.manager.draw_ui(self.surface)

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        """
        Update the state.

        Args:
        ------
        events: The list of pygame events.

        time_delta: The time elapsed since the last update.

        Returns:
        --------
        None: if the state is not changing, Or the the index of the next
        scene.
        """
        for event in events:
            self.manager.process_events(event)
        self.manager.update(time_delta)
        return None


class StateMachine:
    """
    Represents a state machine that manages a collection of states.


    Attributes:
    -----------
    states: The list of states in the state machine.

    state_index: The index of the current active state.
    """

    def __init__(self, states: list[State]):
        """
        Initialize the StateMachine object.

        Args:
        -----
        states: The list of states in the state machine.
        """
        self.states: list[State] = states
        self.state_index: int = 0

    def run_state(self, events: list[pg.Event], time_delta: float):
        """
        Run the current active state in the state machine.

        Args:
        -----
        events: The list of pygame events.

        time_delta: The time elapsed since the last update.
        """
        state: State = self.states[self.state_index]
        state.render()
        new_state_index: Union[int, None] = state.update(events, time_delta)
        self.state_index = (
            new_state_index if new_state_index is not None else self.state_index
        )


class Slider:
    """A class representing a custom UI element"""

    def __init__(
        self,
        label: str,
        position: tuple[int, int],
        initial_value: int,
        slider_range: tuple[int, int],
        manager: pgui.UIManager,
    ):
        """
        Args:
        -----
        label: text to be displayed on the label.

        position: position of the slider.

        initial_value: initial value of the sliding bar.

        slider_range: range of the values that the slider is allowed to get
        into.
        """
        x, y = position
        self.slider = pgui.elements.UIHorizontalSlider(
            pg.Rect(x, y, 400, 30),
            initial_value,
            slider_range,
            manager,
        )
        self.value_label = pgui.elements.UITextBox(
            str(self.slider.get_current_value()),
            pg.Rect(x - 50, y, 50, 60),
            manager,
        )
        self.label = pgui.elements.UITextBox(
            label,
            pg.Rect(x, y + 30, 400, 30),
            manager,
        )

    def update(self):
        """
        Update the value label.
        """
        self.value_label.set_text(str(self.slider.get_current_value()))


class OrganismSelection(State):
    """Represents a organism criteria selection screen."""

    def __init__(
        self,
        surface: pg.Surface,
        world: World,
        next_state_index: Union[int, None],
    ):
        surface_size = width, height = surface.get_size()
        super().__init__(surface, surface_size, next_state_index)
        self.world: World = world

        self.title = pgui.elements.UITextBox(
            "Select how you want to your organisms to initally spawn.",
            pg.Rect((width // 2) - 350 // 2, 50, 350, 70),
            self.manager,
        )

        self.energy_slider_min = Slider(
            "Food min:", (350, 150), 100, (1, 2000), self.manager
        )
        self.energy_slider_max = Slider(
            "Food max:", (350, 250), 1000, (1, 2000), self.manager
        )

        self.temp_slider_min = Slider(
            "Temp min:", (350, 350), 30, (1, 2000), self.manager
        )
        self.temp_slider_max = Slider(
            "Temp max:", (350, 450), 150, (1, 2000), self.manager
        )

        self.done_button = pgui.elements.UIButton(
            pg.Rect(0, 550, -1, -1),
            "Done!",
            self.manager,
            anchors={"centerx": "centerx"},
        )
        self.skip_button = pgui.elements.UIButton(
            pg.Rect(0, 600, -1, -1),
            "skip",
            self.manager,
            anchors={"centerx": "centerx"},
        )
        self.painting_distribution_button = pgui.elements.UIButton(
            pg.Rect(0, 650, -1, -1),
            "customize the other distributions.",
            self.manager,
            anchors={"centerx": "centerx"},
        )

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        """
        Updates the state based on the given events and time delta.

        Args:
        -----
        events: A list of Pygame events.

        time_delta: The time difference between the current and previous frame.
        """
        for event in events:
            if event.type == pgui.UI_BUTTON_PRESSED:
                if event.ui_element == self.done_button:
                    energy_range: tuple[int, int] = (
                        int(self.energy_slider_min.slider.get_current_value()),
                        int(self.energy_slider_max.slider.get_current_value()),
                    )
                    temp_range: tuple[int, int] = (
                        int(self.temp_slider_min.slider.get_current_value()),
                        int(self.temp_slider_max.slider.get_current_value()),
                    )

                    org.Organism.energy_range = energy_range
                    org.Organism.temp_range = temp_range

                    self.world.organism_distribution = (
                        dist.OrganismDistribution.generate(size=self.world.canvas_size)
                    )
                    return self.next_state_index
                if event.ui_element == self.skip_button:
                    return self.next_state_index
                if event.ui_element == self.painting_distribution_button:
                    return -1
            self.manager.process_events(event)

        self.energy_slider_max.update()
        self.energy_slider_min.update()
        self.temp_slider_max.update()
        self.temp_slider_min.update()

        self.manager.update(time_delta)
        return None


class DistributionPainting(State):
    """Represents a distribution painting screen."""

    def __init__(
        self,
        surface: pg.Surface,
        world: World,
        next_state_index: Union[int, None],
    ):
        self.surface_size = width, height = surface.get_size()
        super().__init__(surface, self.surface_size, next_state_index)
        self.world: World = world

        self.canvas_surface: pg.Surface = pg.Surface((500, 500))
        self.canvas_rect: pg.Rect = self.canvas_surface.get_rect(
            center=(width // 2, height // 2)
        )
        self.canvas_offset: Tuple[int, int] = self.canvas_rect.topleft

        self.distribution_labels = {
            self.world.food_distribution: "Food distribution",
            self.world.temp_distribution: "Temperature distribution",
        }

        # State variables
        self.current_distribution: dist.Distribution = self.world.temp_distribution

        # User interface variables
        self.title = pgui.elements.UITextBox(
            "<b>Customize the distributions.</b>",
            pg.Rect((width // 2) - 350 // 2, 20, 350, -1),
            self.manager,
        )
        self.current_distribution_label_text: str = (
            f"<b>Customizing {self.distribution_labels[self.current_distribution]}</b>"
        )

        self.current_distribution_label = pgui.elements.UITextBox(
            self.current_distribution_label_text,
            pg.Rect((width // 2) - 350 // 2, 60, 350, -1),
            self.manager,
        )

        self.instrument: str = "paintbrush"

        self.instrument_indicator = pgui.elements.UITextBox(
            self.instrument, pg.Rect(width - 90, 150, 100, -1), self.manager
        )
        self.paint_button = pgui.elements.UIButton(
            pg.Rect(width - 70, 80, -1, -1), "paint", self.manager
        )
        self.eraser_button = pgui.elements.UIButton(
            pg.Rect(width - 70, 120, -1, -1), "eraser", self.manager
        )

        self.clear_canvas_button = pgui.elements.UIButton(
            pg.Rect(width - 150, 210, -1, -1), "clear the canvas", self.manager
        )
        self.temp_heatmap_button = pgui.elements.UIButton(
            pg.Rect(0, 80, -1, -1), "temp", self.manager
        )
        self.food_heatmap_button = pgui.elements.UIButton(
            pg.Rect(0, 120, -1, -1), "food", self.manager
        )

        self.current_distribution_min_label = pgui.elements.UITextBox(
            f"min: {self.current_distribution.data.min()}",
            pg.Rect(width - 130, 300, -1, -1),
            self.manager,
        )
        self.current_distribution_max_label = pgui.elements.UITextBox(
            f"max: {self.current_distribution.data.max()}",
            pg.Rect(width - 130, 360, -1, -1),
            self.manager,
        )
        self.done_button = pgui.elements.UIButton(
            pg.Rect(0, 650, -1, -1),
            "Done!",
            self.manager,
            anchors={"centerx": "centerx"},
        )

    def render(self):
        render_np_2d_array(self.current_distribution.data, self.canvas_surface)
        self.surface.blit(self.canvas_surface, self.canvas_rect)
        super().render()

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        """
        Updates the state based on the given events and time delta.

        Args:
        -----
        events: A list of Pygame events.

        time_delta: The time difference between the current and previous frame.
        """
        width, height = self.surface_size
        canvas_width, canvas_height = self.canvas_surface.get_size()
        mouse_pos = pg.mouse.get_pos()
        mouse_canvas_pos: Tuple[int, ...] = tuple(
            mouse_pos[_] - self.canvas_offset[_] for _ in range(2)
        )
        mouse_x_rel, mouse_y_rel = mouse_canvas_pos
        is_mouse_in_canvas: bool = (0 < mouse_x_rel < canvas_width) and (
            0 < mouse_y_rel < canvas_height
        )

        if pg.mouse.get_pressed()[0] and is_mouse_in_canvas:
            x_index, y_index = mouse_x_rel // 10, mouse_y_rel // 10
            self.current_distribution.data[y_index][x_index] += (
                100 if self.instrument == "paintbrush" else -100
            )

            self.current_distribution_min_label = pgui.elements.UITextBox(
                f"min: {self.current_distribution.data.min()}",
                pg.Rect(width - 130, 300, -1, -1),
                self.manager,
            )
            self.current_distribution_max_label = pgui.elements.UITextBox(
                f"max: {self.current_distribution.data.max()}",
                pg.Rect(width - 130, 360, -1, -1),
                self.manager,
            )

        for event in events:
            if event.type == pgui.UI_BUTTON_PRESSED:
                if event.ui_element == self.done_button:
                    return self.next_state_index
                if event.ui_element == self.temp_heatmap_button:
                    self.current_distribution = self.world.temp_distribution
                if event.ui_element == self.food_heatmap_button:
                    self.current_distribution = self.world.food_distribution

                if event.ui_element == self.paint_button:
                    self.instrument = "paintbrush"
                if event.ui_element == self.eraser_button:
                    self.instrument = "eraser"

                if event.ui_element == self.clear_canvas_button:
                    self.current_distribution.data = np.zeros(
                        shape=self.current_distribution.data.shape
                    )

                if event.ui_element == self.done_button:
                    return self.next_state_index

                self.instrument_indicator = pgui.elements.UITextBox(
                    self.instrument, pg.Rect(width - 90, 150, 100, -1), self.manager
                )
                self.current_distribution_label_text = (
                    f"<b>Customizing {self.distribution_labels[self.current_distribution]}</b>"
                )  # fmt: skip
                self.current_distribution_label = pgui.elements.UITextBox(
                    self.current_distribution_label_text,
                    pg.Rect((width // 2) - 350 // 2, 60, 350, -1),
                    self.manager,
                )

        return super().update(events, time_delta)


class Simulation(State):
    """Represents the main screen state of the game."""

    def __init__(
        self,
        surface: pg.Surface,
        world: World,
        stats: statistics.StatisticsCollector,
        images: list[pg.Surface],
    ):
        """
        Args:
        -----
        surface: The surface on which the state will be rendered.

        world: The world object containing the simulation data.

        stats: The stats object for collecting data and plotting.

        images: The images representing organisms.
        """

        surface_size = width, height = surface.get_size()
        super().__init__(surface, surface_size, None)

        # Simulation Interface

        self.images: list[pg.Surface] = images

        self.running = False

        self.world: World = world
        self.world_buffer = copy.deepcopy(self.world)
        self.thread = threading.Thread(target=self.world.update_state)
        world_width, world_height = world.canvas_size
        self.world_surface: pg.Surface = pg.surface.Surface(
            (world_height * 64, world_width * 64)
        )
        self.world_rect: pg.Rect = self.world_surface.get_rect(
            center=(width // 2, height // 2)
        )
        self.world_scale: float = 1.0
        self.scaled_world_surface: pg.Surface = self.world_surface

        self.sim_surface: pg.Surface = pg.surface.Surface((width, height))
        self.sim_rect: pg.Rect = self.sim_surface.get_rect(
            center=(width // 2, height // 2)
        )

        self.last_time = 0

        # User Interface
        self.start_button = pgui.elements.UIButton(
            pg.Rect(width - 100, height - 30, 100, 30), "start", self.manager
        )
        self.restart_button = pgui.elements.UIButton(
            pg.Rect(width - 100, height - 60, 100, 30), "restart", self.manager
        )
        self.graph_viz_button = pgui.elements.UIButton(
            pg.Rect(0, 40, -1, -1), "graph", self.manager
        )
        self.temp_heatmap_button = pgui.elements.UIButton(
            pg.Rect(0, 80, -1, -1), "temp", self.manager
        )
        self.food_heatmap_button = pgui.elements.UIButton(
            pg.Rect(0, 120, -1, -1), "food", self.manager
        )
        self.population_label = pgui.elements.UITextBox(
            "0", pg.Rect(0, 0, -1, -1), self.manager
        )
        self.temp_slider = Slider(
            "adjust temperature",
            (width - 500, height - 60),
            45,
            (0, 500),
            self.manager,
        )
        self.food_slider = Slider(
            "adjust the food content",
            (width - 950, height - 60),
            500,
            (0, 1200),
            self.manager,
        )

        # Stats
        self.stats: statistics.StatisticsCollector = stats

    def render(self) -> None:
        """render the main screen state."""
        self.sim_surface.fill("black")
        self.world_surface.fill("#5498C6")
        self.world_buffer.render(self.world_surface, self.images)
        self.sim_surface.blit(self.scaled_world_surface, self.world_rect)
        self.surface.blit(self.sim_surface, self.sim_rect)
        self.manager.draw_ui(self.surface)

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        for event in events:
            if event.type == pgui.UI_BUTTON_PRESSED:
                if event.ui_element == self.start_button:
                    self.running = not self.running
                    self.last_time = 0
                if event.ui_element == self.restart_button:
                    self.running = False
                    self.stats.clear()
                    return 4

                # graph buttons
                if (
                    event.ui_element == self.graph_viz_button
                    and not self.stats.data.empty
                ):
                    self.stats.plot(
                        ["Population", "Food", "Temperature", "Reproduction Ratio"],
                        "Variables plot",
                    )
                if event.ui_element == self.temp_heatmap_button:
                    statistics.plot_heatmap(
                        self.world.temp_distribution.data,
                        "Temperature distribution",
                    )
                if event.ui_element == self.food_heatmap_button:
                    statistics.plot_heatmap(
                        self.world.food_distribution.data,
                        "Food distribution",
                    )
            if event.type == pg.KEYDOWN:
                if event.key == pg.K_SPACE:
                    self.running = not self.running
                    self.last_time = 0

            # change the temp/food content
            if event.type == pgui.UI_HORIZONTAL_SLIDER_MOVED:
                if event.ui_element == self.temp_slider.slider:
                    new_avg_temp = self.temp_slider.slider.get_current_value()
                    self.world.temp_distribution = dist.Distribution.generate(
                        self.world.canvas_size, int(new_avg_temp), 50
                    )
                    self.temp_slider.update()
                if event.ui_element == self.food_slider.slider:
                    new_avg_food_content = self.food_slider.slider.get_current_value()
                    self.world.food_distribution = dist.Distribution.generate(
                        self.world.canvas_size, int(new_avg_food_content), 100
                    )
                    self.food_slider.update()

            self.manager.process_events(event)

        keys_pressed = pg.key.get_pressed()
        # moving
        step_size: int = 500
        if keys_pressed[pg.K_UP] or keys_pressed[pg.K_k]:
            self.world_rect.centery += int(step_size * time_delta)
        if keys_pressed[pg.K_DOWN] or keys_pressed[pg.K_j]:
            self.world_rect.centery -= int(step_size * time_delta)
        if keys_pressed[pg.K_RIGHT] or keys_pressed[pg.K_l]:
            self.world_rect.centerx -= int(step_size * time_delta)
        if keys_pressed[pg.K_LEFT] or keys_pressed[pg.K_h]:
            self.world_rect.centerx += int(step_size * time_delta)

        # can't move beyond
        wiggle_room: int = 100
        if self.world_rect.top > self.sim_rect.top + wiggle_room:
            self.world_rect.top = self.sim_rect.top + wiggle_room
        if self.world_rect.bottom < self.sim_rect.bottom - wiggle_room:
            self.world_rect.bottom = self.sim_rect.bottom - wiggle_room
        if self.world_rect.left > self.sim_rect.left + wiggle_room:
            self.world_rect.left = self.sim_rect.left + wiggle_room
        if self.world_rect.right < self.sim_rect.right - wiggle_room:
            self.world_rect.right = self.sim_rect.right - wiggle_room

        # zooming
        scaling: float = 0.5
        if keys_pressed[pg.K_EQUALS] and self.world_scale < 2:
            self.world_scale += scaling * time_delta
        if keys_pressed[pg.K_MINUS] and self.world_scale > 0.5:
            self.world_scale -= scaling * time_delta

        self.scaled_world_surface = pg.transform.scale_by(
            self.world_surface, self.world_scale
        )
        self.world_rect = self.scaled_world_surface.get_rect(
            center=self.world_rect.center
        )

        if self.running:
            self.start_button.set_text("running")
        else:
            self.start_button.set_text("start")

        # run every 1000 milliseconds
        self.update_sim(1000)

        self.population_label.set_text(
            str(self.world.organism_distribution.get_population())
        )

        self.manager.update(time_delta)
        return None

    def update_sim(self, cycle_time_ms: int):
        """
        for updating the world.

        Args:
        -----
        cycle_time_ms: cycle time under which the world would be updated once
        in milliseconds.
        """
        current_time = pg.time.get_ticks()
        if (
            current_time - self.last_time > cycle_time_ms
            and self.running
            and not self.thread.is_alive()
        ):
            self.world_buffer = copy.deepcopy(self.world)
            self.thread = threading.Thread(target=self.world.update_state)
            self.last_time = current_time
            self.thread.start()
            self.stats.add(
                (
                    self.world.organism_distribution.get_population(),
                    self.world.food_distribution.data.mean(),
                    self.world.temp_distribution.data.mean(),
                    self.world.organism_distribution.get_reproduction_ratio(),
                )
            )


class TitleScreen(State):
    """
    Represents the title screen state of the game.

    Attributes:
    -----------
    font: The font used for rendering the title.

    title_surf: The rendered title surface.

    surface: The surface on which the state is rendered.
    """

    def __init__(
        self,
        surface: pg.Surface,
        title_text: str,
        subtitle_text: str,
        next_state_index: int,
    ) -> None:
        """
        Args:
        -----
        surface: The surface on which the state will be rendered.

        title_text: The text to be displayed as the title.

        subtitle_text: The text to be displayed as the subtitle.

        next_state_index: The index of the next state
        """

        font: pg.Font = pg.font.SysFont("monospace", 25)
        self.title_surf: pg.Surface = font.render(title_text, True, "white")

        smallerfont: pg.Font = pg.font.SysFont("monospace", 12)
        self.subtitle_surf: pg.Surface = smallerfont.render(
            subtitle_text, True, "white"
        )

        self.surface: pg.Surface = surface

        self.next_state_index = next_state_index

    def render(self) -> None:
        """Render the title screen state."""
        self.title_surf = pg.transform.scale(
            self.title_surf,
            (self.surface.get_width(), self.title_surf.get_height()),
        )
        titlerect = self.title_surf.get_rect(center=self.surface.get_rect().center)
        self.surface.blit(self.title_surf, titlerect)

        subtitlerect = self.subtitle_surf.get_rect(
            center=self.surface.get_rect().center
        )
        subtitlerect.centery += 200
        self.surface.blit(self.subtitle_surf, subtitlerect)

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        """
        Update the title screen state.

        Args:
        -----
        events: The list of pygame events.

        time_delta: The time elapsed since the last update.

        Returns:
        -------
        int: The index of the next state to transition to, or None if no
        transition is needed.
        """
        for event in events:
            if event.type == pg.KEYDOWN and event.key != pg.K_F11:
                return self.next_state_index
        return None


class HeadingTextScreen(TitleScreen):
    def __init__(
        self,
        surface: pg.Surface,
        title_text: str,
        content_text: str,
        next_state_index: int,
    ) -> None:
        """
        Args:
        -----
        surface: The surface on which the state will be rendered.

        title_text: The text to be displayed as the title.

        content_text: The text to be displayed as the content.

        next_state_index: The index of the next state
        """

        font: pg.Font = pg.font.SysFont("monospace", 100)
        self.title_surf: pg.Surface = font.render(title_text, True, "white")

        smallerfont: pg.Font = pg.font.SysFont("monospace", 30)
        self.content_text_surf: pg.Surface = smallerfont.render(
            content_text, True, "white"
        )

        self.surface: pg.Surface = surface

        self.next_state_index = next_state_index

    def render(self) -> None:
        """Render the text screen state."""

        titlerect = self.title_surf.get_rect(midtop=self.surface.get_rect().midtop)
        titlerect = self.title_surf.get_rect(midtop=self.surface.get_rect().midtop)
        self.surface.blit(self.title_surf, titlerect)

        subtitlerect = self.content_text_surf.get_rect(
            center=self.surface.get_rect().center
        )
        self.surface.blit(self.content_text_surf, subtitlerect)


class TextScreen(State):
    """
    Represents a screen with text content.

    Attributes:
    -----------
    text_box: The UI text box for displaying the text content.
    """

    def __init__(self, surface: pg.Surface, screen_text: str, next_state_index: int):
        """
        Args:
        -----
        surface: The surface on which the state will be
        rendered.

        screen_text: The text content to be displayed on the screen.

        state_index: The index of the current active state.
        """
        super().__init__(surface, surface.get_size(), next_state_index)
        self.text_box = pgui.elements.UITextBox(
            screen_text, self.surface.get_rect(), self.manager
        )

    def update(self, events: list[pg.Event], time_delta: float) -> Union[int, None]:
        """
        Update the text screen state.

        Args:
        -----
        events: The list of pygame events.

        time_delta: The time elapsed since the last update.

        Returns:
        -------
        int: The index of the next state to transition to, or None if no
        transition is needed.
        """
        for event in events:
            if event.type == pg.KEYDOWN and event.key != pg.K_F11:
                return self.next_state_index
        super().update(events, time_delta)
        return None


def tint(surface: pg.Surface, color: pg.Color) -> pg.Surface:
    """create a tinted surface from the given color"""
    new_surface = surface.copy()
    new_surface.fill(color, special_flags=pg.BLEND_RGB_ADD)
    return new_surface


def get_asset_path(*paths: str):
    """Gets the path for an asset"""
    file_path = files("darwinio")
    file_path = file_path.joinpath("assets")
    for path in paths:
        file_path = file_path.joinpath(path)
    return as_file(file_path)



# The GPLv3 License (GPLv3)

# Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

TITLE_ASCII_ART = """
 ______   _______  _______          _________ _       _________ _______ 
(  __  \ (  ___  )(  ____ )|\     /|\__   __/( (    /|\__   __/(  ___  )
| (  \  )| (   ) || (    )|| )   ( |   ) (   |  \  ( |   ) (   | (   ) |
| |   ) || (___) || (____)|| | _ | |   | |   |   \ | |   | |   | |   | |
| |   | ||  ___  ||     __)| |( )| |   | |   | (\ \) |   | |   | |   | |
| |   ) || (   ) || (\ (   | || || |   | |   | | \   |   | |   | |   | |
| (__/  )| )   ( || ) \ \__| () () |___) (___| )  \  |___) (___| (___) |
(______/ |/     \||/   \__/(_______)\_______/|/    )_)\_______/(_______)
    """

LICENSE_NOTICE = """
The GPLv3 License (GPLv3)

Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/.
"""

HELP = """
To get this help screen, press <b>SHIFT+H</b>.
To zoom in and zoom out, press the <b>-</b> , <b>=</b> keys respectively.
To move around the canvas, use the <b>arrow keys</b> or the vi keys. 
To exit the game, press <b>ESC</b>.
To enter full screen, press <b>F11</b>.
To mute or unmute the music, press <b>m</b>.
"""

DISCLAIMER = """
This program is NOT a game. It is a
simulation that aims to provide a fairly
scientific model of evolution.
"""



# The GPLv3 License (GPLv3)

# Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and related stuff for statistics.
"""
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


class StatisticsCollector:
    """
    A class for collecting and analyzing statistics.

    Attributes:
    ----------
    data: The DataFrame to store the collected statistics.
    """

    def __init__(self, columns: list[str]):
        """
        Initialize a StatisticsCollector object.

        Args:
        -----
        columns: A list of column names for the data.
        """
        self.data: pd.DataFrame = pd.DataFrame(columns=columns)

    def add(self, new_data: tuple[Any, ...]):
        """
        Add new data to the statistics collector.

        Args:
        -----
        new_data: The new data(row) to be added.
        """
        self.data.loc[len(self.data.index)] = new_data

    def plot(self, columns: list[str], title: str):
        """
        Plot the specified columns of the data.

        Args:
        -----
        columns: A list of column names to be plotted.

        title: tile of the plot.
        """
        self.data[columns].plot(subplots=True, title=title, xlabel="time")
        plt.show()

    def clear(self):
        """
        Clear the data in data frame.
        """
        self.data = self.data.iloc[0:0]


def plot_heatmap(data_grid: np.ndarray, title: str):
    """
    Plot the specified columns of the data.

    Args:
    -----
    data_grid: A 2d array of values to be plotted.

    title: tile of the plot.
    """
    plt.imshow(data_grid, interpolation="nearest")
    plt.title(title)
    plt.show()



# The GPLv3 License (GPLv3)

# Copyright © 2023 Tushar Maharana, and Mihir Nallagonda

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Classes and methods for handling the neural network.

Classes:
-------------
NeuralNetwork: A class for representing a neural network with a given genome
and neural structure.

Functions:
-------------
create_weights: Creates the weights for the neural network based on the
genome and neural structure.

normalize_array: Creates the weights for the neural network based on the genome
and neural structure.
"""

from __future__ import annotations

import numpy as np


class NeuralNetwork:
    """
    A class for representing a neural network with a given genome and neural
    structure.

    Attributes:
    ----------
    weights: contains weights of the neural network.

    neural_structure: A numpy ndarray of integers representing the structure of
    the neural network, where each integer represents the number of neurons in
    that layer.
    """

    def __init__(self, weights: np.ndarray, neural_structure: np.ndarray) -> None:
        """
        Initialize a neural network object with the given weight and neural
        structure.

        Args:
        -----
        weights: An one-dimensional Numpy Array containing the weights for
        the neural network.

        neural_structure: An one-dimensional Numpy Array representing the
        neural structure of the organism.
        """

        self.weights: np.ndarray = weights
        self.neural_structure: np.ndarray = neural_structure

    def run_neural_network(self, input_values: np.ndarray) -> np.ndarray:
        """
        Run the neural network according to input values.

        Args:
        -----
        input_values: A numpy ndarray representing the input values of the
        neural network.

        Returns:
        --------
        A numpy ndarray with the output values of the neural network

        Note:
        -----
        This method runs the neural network with the given input values using
        the current weights and neural structure of the neural network. It
        initializes a new empty neural network and fills in the values using
        the provided input and the weights. The weights are then applied to the
        corresponding input neuron to obtain the output of the next neuron in
        the network. The output of each layer is passed through the activation
        function (tanh) and stored in the neural network list. Finally, the
        output of the last layer is returned as a Numpy Array.
        """

        normalized_input_values: np.ndarray = normalize(input_values)
        weights: np.ndarray = self.weights
        neural_structure: np.ndarray = self.neural_structure
        neural_network: list[list[float]] = [[0.0] * i for i in neural_structure]

        neural_network[0] = list(normalized_input_values)

        for layer_index in range(len(neural_network[:-1])):
            layer_values = neural_network[layer_index]
            next_layer_index = layer_index + 1

            # calculate the dot product between the current layer and the
            # weights for the next layer
            try:
                next_layer_values = np.dot(
                    layer_values,
                    weights[
                        : len(layer_values) * len(neural_network[next_layer_index])
                    ].reshape(len(layer_values), len(neural_network[next_layer_index])),
                )
            except:
                next_layer_values = np.dot(layer_values, np.ones((2, 2)))

            # reshape the resulting values to match the shape of the next layer
            next_layer_values = next_layer_values.reshape(
                len(neural_network[next_layer_index]),
            )

            # apply the activation function to the next layer values
            next_layer_values = np.tanh(next_layer_values)
            neural_network[next_layer_index] = list(next_layer_values)

        return np.array(neural_network[-1])


def normalize(arr: np.ndarray) -> np.ndarray:
    """Normalize the Numpy Array."""
    magnitude: np.floating = np.linalg.norm(arr) + 1e-16
    return arr / magnitude


def create_weights(
    genome_array: np.ndarray, neural_structure: np.ndarray
) -> np.ndarray:
    """
    Creates the weights for the neural network based on the genome and
    neural structure.

    Args:
    -----
    genome_array: A Numpy Array representing the hexadecimal genome of the
    organism.

    neural_structure: A Numpy Array representing the structure of the neural
    network.

    Returns:
    -------
    A Numpy Array containing the weights.
    """

    number_of_neural_connections = int(
        np.sum(neural_structure[:-1] * neural_structure[1:])
    )

    # generate the weights
    weights: np.ndarray = genome_array[:number_of_neural_connections]

    # normalize to the range [-1, 1]
    weights: np.ndarray = normalize(weights)

    return weights
